"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.TestCompromisedError = void 0;
const ErrorSerialiser_1 = require("../ErrorSerialiser");
const RuntimeError_1 = require("./RuntimeError");
/**
 * Thrown to indicate that the test can't be performed due to an issue with a downstream dependency.
 * For example, it makes no sense to run a full-stack integration test if we already know that
 * the database server is down.
 *
 * ## Throwing a TestCompromisedError from a custom Interaction
 *
 * ```ts
 * import { Interaction } from '@serenity-js/core';
 *
 * const SetUpTestDatabase = () =>
 *   Interaction.where(`#actor sets up a test database`, actor => {
 *     return SomeCustomDatabaseSpecificAbility.as(actor).setUpTestDatabase().catch(error => {
 *       throw new TestCompromisedError('Could not set up the test database', error)
 *     })
 * })
 * ```
 *
 * @group Errors
 */
class TestCompromisedError extends RuntimeError_1.RuntimeError {
    static fromJSON(serialised) {
        const error = new TestCompromisedError(serialised.message, ErrorSerialiser_1.ErrorSerialiser.deserialise(serialised.cause));
        error.stack = serialised.stack;
        return error;
    }
    /**
     * @param message - Human-readable description of the error
     * @param [cause] - The root cause of this [`RuntimeError`](https://serenity-js.org/api/core/class/RuntimeError/), if any
     */
    constructor(message, cause) {
        super(TestCompromisedError, message, cause);
    }
}
exports.TestCompromisedError = TestCompromisedError;
//# sourceMappingURL=TestCompromisedError.js.map