import type { ListensToDomainEvents } from './ListensToDomainEvents';
import type { Stage } from './Stage';

/**
 * You can think of the [`StageCrewMember`](https://serenity-js.org/api/core/interface/StageCrewMember/) as an in-memory microservice that reacts to [Serenity/JS domain events](https://serenity-js.org/api/core-events/class/DomainEvent/),
 * published by the [`StageManager`](https://serenity-js.org/api/core/class/StageManager/), and originally emitted by [actors](https://serenity-js.org/api/core/class/Actor/) performing activities and Serenity/JS
 * test runner adapters notifying the framework about test scenario-specific events like [SceneStarts](https://serenity-js.org/api/core-events/class/SceneStarts/) or [SceneFinishes](https://serenity-js.org/api/core-events/class/SceneFinishes/).
 *
 * Every [stage crew member](https://serenity-js.org/api/core/interface/StageCrewMember/) receives a reference to the [stage](https://serenity-js.org/api/core/class/Stage/),
 * and therefore the [stage manager](https://serenity-js.org/api/core/class/StageManager/), which enables them to emit [domain events](https://serenity-js.org/api/core-events/class/DomainEvent/) back.
 *
 * This interface is useful when you're interested in implementing custom Serenity/JS reporters or supporting services.
 *
 * ## Implementing a custom StageCrewMember
 *
 * ```ts
 * import { Stage, StageCrewMember } from '@serenity-js/core'
 * import * as events from '@serenity-js/core/lib/events'
 * import { ArbitraryTag } from '@serenity-js/core/lib/model'
 *
 * export class TestRunnerTagger implements StageCrewMember {
 *
 *   static using(tagName: string) {
 *     return new TestRunnerTagger(tagName);
 *   }
 *
 *   protected constructor(
 *     private readonly tagName: string,
 *     private stage?: Stage,
 *   ) {
 *   }
 *
 *   assignedTo(stage: Stage): StageCrewMember {
 *     this.stage = stage;
 *     return this;
 *   }
 *
 *   notifyOf(event: events.DomainEvent): void {
 *     if (event instanceof events.TestRunnerDetected) {
 *       this.stage.announce(
 *         new events.SceneTagged(
 *           this.stage.currentSceneId(),
 *           new ArbitraryTag(this.tagName),
 *           this.stage.currentTime()
 *         )
 *       )
 *     }
 *   }
 * }
 * ```
 *
 * ## Using the custom StageCrewMember
 *
 * ```ts
 * import { configure } from '@serenity-js/core'
 *
 * configure({
 *   crew: [
 *     TestRunnerTagger.using(`Node:${ process.version }`),
 *   ]
 * })
 * ```
 *
 * ## Learn more
 * - [`SerenityConfig.crew`](https://serenity-js.org/api/core/class/SerenityConfig/#crew)
 * - [`configure`](https://serenity-js.org/api/core/function/configure/)
 * - [`StageCrewMember`](https://serenity-js.org/api/core/interface/StageCrewMemberBuilder/)
 * - [`ListensToDomainEvents`](https://serenity-js.org/api/core/interface/ListensToDomainEvents/)
 * - [`DomainEvent`](https://serenity-js.org/api/core-events/class/DomainEvent/)
 *
 * @group Stage
 */
export interface StageCrewMember extends ListensToDomainEvents {

    /**
     * Assigns a [`Stage`](https://serenity-js.org/api/core/class/Stage/) to this [`StageCrewMember`](https://serenity-js.org/api/core/interface/StageCrewMember/)
     */
    assignedTo(stage: Stage): StageCrewMember;
}
